

DEFAULT_BROM_DIR = "./brom" #: Default directory for BROM Drake data storage
DEFAULT_WATCHER_DIR = DEFAULT_BROM_DIR + "/watcher" #: Default subdirectory for :py:class:`PortWatcher<brom_drake.PortWatcher.port_watcher.PortWatcher>` data
DEFAULT_PLOT_DIR = DEFAULT_WATCHER_DIR + "/plots" #: Default subdirectory for :py:class:`PortWatcher<brom_drake.PortWatcher.port_watcher.PortWatcher>` plots
DEFAULT_RAW_DATA_DIR = DEFAULT_WATCHER_DIR + "/raw_data" #: Default subdirectory for :py:class:`PortWatcher<brom_drake.PortWatcher.port_watcher.PortWatcher>` raw data
DEFAULT_BROM_MODELS_DIR = DEFAULT_BROM_DIR + "/models" #: Default subdirectory for BROM Drake models
DEFAULT_NETWORKX_FSM_DIR = DEFAULT_BROM_DIR + "/networkx_fsm" #: Default subdirectory for NetworkX FSM data